#include <sys/types.h>
#include <dirent.h>
#import <Foundation/NSString.h>
#import "FolderMoviesPatch.h"
#import "QCStructure.h"

@implementation FolderMoviesPatch : QCPatch

+ (int)executionMode
{
	return 2;
}
+ (BOOL)allowsSubpatches
{
	return NO;
}
+ (int)timeMode
{
	return 1;
}

// scan thread method
+(void) scanThread:(id)object
{
	/*
		Some conventions for this thread:  We have 2 dictionaries/structures.  This is to mimic how Folder Images 
		Operates.  We have a root structure (denoted as "full*" in this code) which containst sub-structures.  The
		Keys for these structures are NSStrings in the form of "file_N" where N starts at 1 and counts up.
		
		This function really only deals with the "full*" structures, so it's easy to grasp.
		
		For reference: our root dictionary is just a QCStructure for now.  It can work as a dictionary too, 
		but the indexes get shuffled around for some reason.  Need to investigate this further.
	*/
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	FolderMoviesPatch *p = (FolderMoviesPatch*)object;
	//NSMutableDictionary *fullDict = [[NSMutableDictionary alloc] init];
	
	// scan to populate the full structure
	//NSLog(@"FM Scanning...\n");
	QCStructure *fullStruct = [[QCStructure alloc] init];
	[self scanPath:[p->inputPath stringValue] withDict:fullStruct recursively:[p->inputRecursive booleanValue]];
	//NSLog(@"FM Scan Complete\n");
	
	// here we populate the full structure with our scanned dictionary.
	//NSLog(@"FM dict: %s\n",[[fullDict description] UTF8String]);
	// initWithDictionary seems to shuffle the index order.  we don't do this now, but it's the preferred way
	// (tooltip data is formatted nicely when initialized with a dict) for 1.0  
	//QCStructure *fullStruct = [[QCStructure alloc] initWithDictionary:fullDict];

	[p->outputFileList setStructureValue:fullStruct];

	[p->outputCompleteSignal setBooleanValue:TRUE];
	[pool release];
	
	p->currentlyScanning = FALSE;
}

/*
	Continuing on our sub-structure explanation, in this function (the sometimes-recursive scanPath function) we handle the
	"full*" object only to add sub-structures.  The sub-structures are denoted with "file*", indicating their one-to-one
	pairing with an actual on-disk file.
	
	There are probably a few ways to do this.  I'm using stdlib stuff for now, but we might need
	to break out a pure Cocoa version to handle exotic (non-ASCII) filenames perhaps?
*/
+ (void)scanPath:(NSString*)path withDict:(QCStructure*)fullDict recursively:(BOOL) recursive
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	DIR *pathdir;
	struct dirent *entry;
	//NSLog(@"FM Scanning '%s'\n",[path UTF8String]);
	
	pathdir = opendir([path UTF8String]);

	while (( entry = readdir(pathdir)) != NULL)
	{
		//NSLog(@"FM Inspecting %s\n",entry->d_name);
		if( entry->d_namlen > 4)
		{
			// search for regular files only.  No symlinks.
			if( entry->d_type == DT_REG )
			{
				/* UTI code shamelessly stolen from KinemeFileType patch by smokris. */
				NSURL *filepathURL;
				if([path characterAtIndex:[path length]-1] == '/')
					filepathURL = [NSURL fileURLWithPath:[NSString stringWithFormat:@"%s%s",[path UTF8String],entry->d_name]];
				else
					filepathURL = [NSURL fileURLWithPath:[NSString stringWithFormat:@"%s/%s",[path UTF8String],entry->d_name]];
				FSRef	ref;
				LSItemInfoRecord outInfo;
				CFURLRef cfURLRef = (CFURLRef)filepathURL;
				
				if( CFURLGetFSRef(cfURLRef, &ref) == TRUE)
				{
					outInfo.extension = NULL;
					if(LSCopyItemInfoForRef(&ref,kLSRequestExtension|kLSRequestTypeCreator, &outInfo) == noErr)
					{
						CFStringRef itemUTI = NULL;
						if(outInfo.extension != NULL)
						{
							itemUTI = UTTypeCreatePreferredIdentifierForTag 
										(kUTTagClassFilenameExtension,outInfo.extension, NULL);
							CFRelease(outInfo.extension);
						}
						else
						{
                            CFStringRef typeString = UTCreateStringForOSType(outInfo.filetype);
							itemUTI = UTTypeCreatePreferredIdentifierForTag(kUTTagClassFilenameExtension, typeString, NULL);
							CFRelease( typeString );
						}
						
						if(itemUTI != NULL)
						{
							// we're looking for "public.movie" itemUTIs to add to the list.
							//NSLog(@"FM uti for %s: %s\n",[[filepathURL absoluteString] UTF8String],[(NSString*)itemUTI UTF8String]);
							if( UTTypeConformsTo(itemUTI,(CFStringRef)@"public.movie") == TRUE )
							{
								NSMutableDictionary *fileDict = [[NSMutableDictionary alloc] init];
								[fileDict
									setObject:[NSString stringWithFormat:@"%s",entry->d_name]
									forKey:@"name"];
								// here we check to see if the path has a trailing slash.  if so, we don't need to add one.
								if([path characterAtIndex:[path length]-1] == '/')
									[fileDict 
										setObject:[NSString stringWithFormat:@"%s%s",[path UTF8String],entry->d_name] 
										forKey:@"path"];
								else
									[fileDict 
										setObject:[NSString stringWithFormat:@"%s/%s",[path UTF8String],entry->d_name] 
										forKey:@"path"];
								NSURL *fileURL = [NSURL fileURLWithPath:[fileDict objectForKey:@"path"]];
								[fileDict
									setObject:[fileURL absoluteString]
									forKey:@"url"];
									
								[fullDict setMember:fileDict forKey:[NSString stringWithFormat:@"file_%i",[fullDict count]+1] ];
							}
						}
					}
				}
			}
		}
	}

	// now, we need to walk the directories if we're in recursive mode
	if(recursive == TRUE)
	{
		rewinddir(pathdir);
		while (( entry = readdir(pathdir)) != NULL)
		{
			// . and .. don't count
			if(entry->d_type == DT_DIR && strcmp("..",entry->d_name) != 0 && strcmp(".",entry->d_name) != 0)
			{
				//NSLog(@"recursing %s\n",[[NSString stringWithFormat:@"%s/%s",[path UTF8String],entry->d_name] UTF8String]);
				if([path characterAtIndex:[path length]-1] == '/')
					[self scanPath:[NSString stringWithFormat:@"%s%s",[path UTF8String],entry->d_name] withDict:fullDict recursively:recursive];
				else
					[self scanPath:[NSString stringWithFormat:@"%s/%s",[path UTF8String],entry->d_name] withDict:fullDict recursively:recursive];
			}
		}
	}

	closedir(pathdir);
	[pool release];
}

- (id)initWithIdentifier:(id)fp8
{
	id z=[super initWithIdentifier:fp8];
	if(z)
	{
		oldValue = FALSE;
		currentlyScanning = FALSE;
	}
	return z;
}


- (id)setup:(id)fp8
{
	return fp8;
}
- (void)cleanup:(id)fp8
{
}


- (void)enable:(id)fp8
{
}
- (void)disable:(id)fp8
{
}


- (BOOL)execute:(id)fp8 time:(double)fp12 arguments:(id)fp20
{
	// if signaled and we're not currently scanning...
	if([inputScanSignal booleanValue] == TRUE && currentlyScanning == FALSE && oldValue == FALSE)
	{
		// ... further check: if path is empty, don't bother creating a thread and all that...
		oldValue = TRUE;
		if([[inputPath stringValue] length] > 0)
		{
			// ... we're clear to begin scanning
			//NSLog(@"FM: Scan Thread Engaged\n");
			currentlyScanning = TRUE;
			[outputCompleteSignal setBooleanValue:FALSE];
			[NSThread detachNewThreadSelector:@selector(scanThread:) toTarget:[FolderMoviesPatch class] withObject:self];
		}
	}
	else
		oldValue = [inputScanSignal booleanValue];
		
	return YES;
}

@end
